@@ -11,6 +11,8 @@ import morecantile
 
 logger = logging.getLogger(__name__)
 
+WGS84_CRS = CRS.from_epsg(4326)
+
 
 def configure_logging(verbosity):
     """Configure log verbosity.
@@ -53,15 +55,15 @@ def normalize_source(input):
     first_line = next(src)
 
     # If input is RS-delimited JSON sequence.
-    if first_line.startswith(u"\x1e"):
+    if first_line.startswith("\x1e"):
 
         def feature_gen():
-            buffer = first_line.strip(u"\x1e")
+            buffer = first_line.strip("\x1e")
             for line in src:
-                if line.startswith(u"\x1e"):
+                if line.startswith("\x1e"):
                     if buffer:
                         yield json.loads(buffer)
-                    buffer = line.strip(u"\x1e")
+                    buffer = line.strip("\x1e")
                 else:
                     buffer += line
             else:
@@ -173,7 +175,7 @@ def cli(ctx, verbose, quiet):
 @click.option(
     "--extents/--no-extents",
     default=False,
-    help="Write shape extents as ws-separated strings (default is " "False).",
+    help="Write shape extents as ws-separated strings (default is False).",
 )
 # Optionally buffer the shapes by shifting the x and y values of each
 # vertex by a constant number of decimal degrees or meters (depending
@@ -184,8 +186,18 @@ def cli(ctx, verbose, quiet):
     default=None,
     help="Shift shape x and y values by a constant number",
 )
+@click.option(
+    "--tms",
+    help="Path to TileMatrixSet JSON file.",
+    type=click.Path(),
+)
+@click.option(
+    "--crs",
+    help="Geographic CRS. Default to WGS84.",
+    type=str,
+)
 @click.pass_context
-def shapes(
+def shapes(  # noqa: C901
     ctx,
     input,
     identifier,
@@ -198,6 +210,8 @@ def shapes(
     collect,
     extents,
     buffer,
+    tms,
+    crs,
 ):
     """
     Reads one or more Web Mercator tile descriptions
@@ -217,7 +231,10 @@ def shapes(
     the properties object of the output feature.
 
     """
-    tms = morecantile.tms.get(identifier)
+    tilematrixset = morecantile.tms.get(identifier)
+    if tms:
+        with open(tms, "r") as f:
+            tilematrixset = morecantile.TileMatrixSet(**json.load(f))
 
     dump_kwds = {"sort_keys": True}
     if indent:
@@ -230,7 +247,7 @@ def shapes(
     col_xs = []
     col_ys = []
 
-    for i, line in enumerate(iter_lines(src)):
+    for _i, line in enumerate(iter_lines(src)):
         obj = json.loads(line)
         if isinstance(obj, dict):
             x, y, z = obj["tile"][:3]
@@ -243,13 +260,14 @@ def shapes(
         else:
             raise click.BadParameter("{0}".format(obj), param=input, param_hint="input")
 
-        feature = tms.feature(
+        feature = tilematrixset.feature(
             (x, y, z),
             fid=fid,
             props=props,
             projected=projected,
             buffer=buffer,
             precision=precision,
+            geographic_crs=CRS.from_user_input(crs) if crs else WGS84_CRS,
         )
         bbox = feature["bbox"]
         w, s, e, n = bbox
@@ -262,7 +280,7 @@ def shapes(
             click.echo(" ".join(map(str, bbox)))
         else:
             if seq:
-                click.echo(u"\x1e")
+                click.echo("\x1e")
             if output_mode == "bbox":
                 click.echo(json.dumps(bbox, **dump_kwds))
             elif output_mode == "feature":
@@ -273,7 +291,7 @@ def shapes(
         click.echo(
             json.dumps(
                 {"type": "FeatureCollection", "bbox": bbox, "features": features},
-                **dump_kwds
+                **dump_kwds,
             )
         )
 
@@ -302,8 +320,13 @@ def shapes(
     default=False,
     help="Write a RS-delimited JSON sequence (default is LF).",
 )
+@click.option(
+    "--tms",
+    help="Path to TileMatrixSet JSON file.",
+    type=click.Path(),
+)
 @click.pass_context
-def tiles(ctx, zoom, input, identifier, seq):
+def tiles(ctx, zoom, input, identifier, seq, tms):  # noqa: C901
     """
     Lists TMS tiles at ZOOM level intersecting
     GeoJSON [west, south, east, north] bounding boxen, features, or
@@ -324,7 +347,10 @@ def tiles(ctx, zoom, input, identifier, seq):
     [853, 1551, 12]
 
     """
-    tms = morecantile.tms.get(identifier)
+    tilematrixset = morecantile.tms.get(identifier)
+    if tms:
+        with open(tms, "r") as f:
+            tilematrixset = morecantile.TileMatrixSet(**json.load(f))
 
     for obj in normalize_source(input):
         if isinstance(obj, list):
@@ -362,11 +388,13 @@ def tiles(ctx, zoom, input, identifier, seq):
             east -= epsilon
             north -= epsilon
 
-        for tile in tms.tiles(west, south, east, north, [zoom], truncate=False):
+        for tile in tilematrixset.tiles(
+            west, south, east, north, [zoom], truncate=False
+        ):
             vals = (tile.x, tile.y, zoom)
             output = json.dumps(vals)
             if seq:
-                click.echo(u"\x1e")
+                click.echo("\x1e")
 
             click.echo(output)
 
@@ -383,7 +411,7 @@ def tiles(ctx, zoom, input, identifier, seq):
 def tms(identifier):
     """Print TMS JSON."""
     tms = morecantile.tms.get(identifier)
-    click.echo(tms.json(exclude_none=True))
+    click.echo(tms.model_dump_json(exclude_none=True))
 
 
 ################################################################################
@@ -408,7 +436,7 @@ def tms(identifier):
     help="Identifier of the custom TMS.",
     default="CustomTileMatrixSet",
 )
-@click.option("--minzoom", type=int, default=0, help="Minumum Zoom level.")
+@click.option("--minzoom", type=int, default=0, help="Minimum Zoom level.")
 @click.option("--maxzoom", type=int, default=24, help="Maximum Zoom level.")
 @click.option("--tile-width", type=int, default=256, help="Width of each tile.")
 @click.option("--tile-height", type=int, default=256, help="Height of each tile.")
@@ -417,21 +445,27 @@ def tms(identifier):
     type=int,
     help="EPSG number for the bounding box.",
 )
-def custom(epsg, extent, name, minzoom, maxzoom, tile_width, tile_height, extent_epsg):
+@click.option(
+    "--title",
+    type=str,
+    help="Tile Matrix Set title.",
+)
+def custom(
+    epsg, extent, name, minzoom, maxzoom, tile_width, tile_height, extent_epsg, title
+):
     """Create Custom TMS."""
-    extent_crs = CRS.from_epsg(extent_epsg) if extent_epsg else None
-
     tms = morecantile.TileMatrixSet.custom(
         extent,
         CRS.from_epsg(epsg),
-        identifier=name,
+        id=name,
         minzoom=minzoom,
         maxzoom=maxzoom,
         tile_width=tile_width,
         tile_height=tile_height,
-        extent_crs=extent_crs,
+        extent_crs=CRS.from_epsg(extent_epsg) if extent_epsg else None,
+        title=title or "Custom TileMatrixSet",
     )
-    click.echo(tms.json(exclude_none=True))
+    click.echo(tms.model_dump_json(exclude_none=True))
 
 
 ################################################################################
@@ -488,7 +522,7 @@ def custom(epsg, extent, name, minzoom, maxzoom, tile_width, tile_height, extent
 @click.option(
     "--extents/--no-extents",
     default=False,
-    help="Write shape extents as ws-separated strings (default is " "False).",
+    help="Write shape extents as ws-separated strings (default is False).",
 )
 # Optionally buffer the shapes by shifting the x and y values of each
 # vertex by a constant number of decimal degrees or meters (depending
@@ -499,7 +533,12 @@ def custom(epsg, extent, name, minzoom, maxzoom, tile_width, tile_height, extent
     default=None,
     help="Shift shape x and y values by a constant number",
 )
-def tms_to_geojson(
+@click.option(
+    "--crs",
+    help="Geographic CRS. Default to WGS84.",
+    type=str,
+)
+def tms_to_geojson(  # noqa: C901
     input,
     level,
     precision,
@@ -511,6 +550,7 @@ def tms_to_geojson(
     collect,
     extents,
     buffer,
+    crs,
 ):
     """Print TMS document as GeoJSON."""
     tms = morecantile.TileMatrixSet(**json.load(input))
@@ -526,13 +566,22 @@ def tms_to_geojson(
     col_xs = []
     col_ys = []
 
-    for x in range(0, matrix.matrixWidth):
-        for y in range(0, matrix.matrixHeight):
+    for y in range(0, matrix.matrixHeight):
+        cf = (
+            matrix.get_coalesce_factor(y)
+            if matrix.variableMatrixWidths is not None
+            else 1
+        )
+        for x in range(0, matrix.matrixWidth):
+            if cf != 1 and x % cf:
+                continue
+
             feature = tms.feature(
                 (x, y, level),
                 projected=projected,
                 buffer=buffer,
                 precision=precision,
+                geographic_crs=CRS.from_user_input(crs) if crs else WGS84_CRS,
             )
             bbox = feature["bbox"]
             w, s, e, n = bbox
@@ -545,7 +594,7 @@ def tms_to_geojson(
                 click.echo(" ".join(map(str, bbox)))
             else:
                 if seq:
-                    click.echo(u"\x1e")
+                    click.echo("\x1e")
                 if output_mode == "bbox":
                     click.echo(json.dumps(bbox, **dump_kwds))
                 elif output_mode == "feature":
